from __future__ import annotations

from bfabric.entities.core.relationship import Relationship
from bfabric.entities.user import User
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import TypeVar

    T = TypeVar("T")


class HasUser(Relationship[User]):
    def __init__(self, bfabric_field: str, *, optional: bool = False) -> None:
        super().__init__(entity="User")
        self._bfabric_field = bfabric_field
        self._optional = optional

    def __get__(self, obj: T | None, objtype: type[T] | None = None) -> User | None:
        cache_attr = f"_HasUser__{self._bfabric_field}_cache"
        if not hasattr(obj, cache_attr):
            setattr(obj, cache_attr, self._load_user(obj=obj))
        return getattr(obj, cache_attr)

    def _load_user(self, obj: T) -> User | None:
        from bfabric.entities.user import User

        client = obj._client
        user = User.find_by_login(login=obj.data_dict[self._bfabric_field], client=client)
        if user is not None or self._optional:
            return user
        else:
            raise ValueError(f"Field {repr(self._bfabric_field)} is required")



from __future__ import annotations

from typing import TypeVar, TYPE_CHECKING

from loguru import logger

from bfabric.experimental.cache._fifo_cache import FifoCache

if TYPE_CHECKING:
    from bfabric.entities.core.entity import Entity

    E = TypeVar("E", bound="Entity")


class EntityMemoryCache:
    """Implements a configurable cache for different entity types in memory.

    Only entity types specified in the config will be cached, and at most the specified number of entities will be
    cached for each type.
    """

    def __init__(self, config: dict[type[Entity], int]) -> None:
        self._config = config
        self._caches = {entity_type: FifoCache(max_size=max_size) for entity_type, max_size in config.items()}

    def contains(self, entity_type: type[Entity], entity_id: int) -> bool:
        """Returns whether the cache contains an entity with the given type and ID."""
        if entity_type not in self._caches:
            return False
        else:
            return entity_id in self._caches[entity_type]

    def get(self, entity_type: type[E], entity_id: int) -> E | None:
        """Returns the entity with the given type and ID, if it exists in the cache."""
        if entity_type not in self._caches:
            return None

        if self._caches[entity_type].get(entity_id):
            logger.debug(f"Cache hit for entity {entity_type} with ID {entity_id}")
            return self._caches[entity_type].get(entity_id)
        else:
            logger.debug(f"Cache miss for entity {entity_type} with ID {entity_id}")
            return None

    def get_all(self, entity_type: type[Entity], entity_ids: list[int]) -> dict[int, Entity]:
        """Returns a dictionary of entities with the given type and IDs,
        containing only the entities that exist in the cache.
        """
        results = {entity_id: self.get(entity_type, entity_id) for entity_id in entity_ids}
        return {entity_id: result for entity_id, result in results.items() if result is not None}

    def put(self, entity_type: type[Entity], entity_id: int, entity: Entity | None) -> None:
        """Puts an entity with the given type and ID into the cache."""
        if entity_type not in self._caches:
            return

        logger.debug(f"Caching entity {entity_type} with ID {entity_id}")
        self._caches[entity_type].put(entity_id, entity)



from __future__ import annotations
from typing import Any, TYPE_CHECKING

from bfabric.entities.core.entity import Entity

if TYPE_CHECKING:
    from bfabric import Bfabric


class User(Entity):
    ENDPOINT = "user"

    def __init__(self, data_dict: dict[str, Any], client: Bfabric | None = None) -> None:
        super().__init__(data_dict=data_dict, client=client)

    @classmethod
    def find_by_login(cls, login: str, client: Bfabric) -> User | None:
        """Finds a user by their login name."""
        users = cls.find_by({"login": login}, client=client)
        if not users:
            return None
        return list(users.values())[0]



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bfabric.entities.core.entity import Entity
    from bfabric.experimental.cache._entity_memory_cache import EntityMemoryCache


class CacheStack:
    """Implements a stack of entity caches.

    Retrieval is performed by looking first at the most recently added cache.
    Items are put into all caches in the stack (they may each have different sizes and inclusion rules) so that nesting
    contexts has no effect on cache hits.
    """

    def __init__(self) -> None:
        self._stack: list[EntityMemoryCache] = []

    def cache_push(self, cache: EntityMemoryCache) -> None:
        self._stack.append(cache)

    def cache_pop(self) -> None:
        self._stack.pop()

    def item_contains(self, entity_type: type[Entity], entity_id: int) -> bool:
        return any(cache.contains(entity_type, entity_id) for cache in reversed(self._stack))

    def item_get(self, entity_type: type[Entity], entity_id: int) -> Entity | None:
        for cache in reversed(self._stack):
            entity = cache.get(entity_type, entity_id)
            if entity is not None:
                return entity
        return None

    def item_get_all(self, entity_type: type[Entity], entity_ids: list[int]) -> dict[int, Entity]:
        results = {}
        pending = set(entity_ids)
        for cache in reversed(self._stack):
            if not pending:
                break
            matches = cache.get_all(entity_type, list(pending))
            results.update(matches)
            pending.difference_update(matches)
        return results

    def item_put(self, entity_type: type[Entity], entity_id: int, entity: Entity | None) -> None:
        for cache in reversed(self._stack):
            cache.put(entity_type, entity_id, entity)

    def item_put_all(self, entity_type: type[Entity], entities: dict[int, Entity | None]) -> None:
        for entity_id, entity in entities.items():
            self.item_put(entity_type, entity_id, entity)



from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING, overload

from bfabric.experimental.cache._cache_stack import CacheStack
from bfabric.experimental.cache._entity_memory_cache import EntityMemoryCache

if TYPE_CHECKING:
    from bfabric.entities.core.entity import Entity
    from collections.abc import Iterator


@overload
def cache_entities(entities: type[Entity], max_size: int) -> Iterator[None]: ...


@overload
def cache_entities(entities: list[type[Entity]], max_size: int) -> Iterator[None]: ...


@overload
def cache_entities(entities: dict[type[Entity], int]) -> Iterator[None]: ...


@contextmanager
def cache_entities(
    entities: type[Entity] | list[type[Entity]] | dict[type[Entity], int], max_size: int = 0
) -> Iterator[None]:
    """Enables caching for the specified entity types within the context."""
    # Get the entities config dict
    if isinstance(entities, dict):
        config = entities
    elif isinstance(entities, list):
        config = {entity: max_size for entity in entities}
    else:
        config = {entities: max_size}

    # Get the stack
    stack = get_cache_stack()

    # Create the new cache and push it to the stack
    cache = EntityMemoryCache(config=config)
    stack.cache_push(cache)

    # Yield to the context
    try:
        yield
    finally:
        stack.cache_pop()


# Context variable to store the cache stack per async task/thread context
_cache_stack_var: ContextVar[CacheStack | None] = ContextVar("cache_stack", default=None)


def get_cache_stack() -> CacheStack:
    """Returns the cache stack instance for the current context."""
    cache = _cache_stack_var.get()
    if cache is None:
        cache = CacheStack()
        _cache_stack_var.set(cache)
    return cache


def _reset_cache_stack() -> None:
    """Reset the cache stack for the current context (for testing)."""
    _cache_stack_var.set(None)



from __future__ import annotations

from collections import OrderedDict
from typing import TypeVar, Generic, TYPE_CHECKING

T = TypeVar("T")

if TYPE_CHECKING:
    from collections.abc import Hashable


class FifoCache(Generic[T]):
    """A FIFO cache with a maximum size, implemented by an OrderedDict."""

    def __init__(self, max_size: int) -> None:
        self._entries: OrderedDict[Hashable, T] = OrderedDict()
        self._max_size = max_size

    def get(self, key: Hashable) -> T | None:
        """Returns the value with the given key, if it exists, and marks it as used.

        If the key does not exist, returns None.
        """
        if key in self._entries:
            self._entries.move_to_end(key)
            return self._entries[key]

    def put(self, key: Hashable, value: T) -> None:
        """Puts a key-value pair into the cache, marking it as used."""
        if self._max_size != 0 and len(self._entries) >= self._max_size:
            self._entries.popitem(last=False)
        self._entries[key] = value

    def __contains__(self, key: Hashable) -> bool:
        """Returns whether the cache contains a key."""
        return key in self._entries



from .application import Application
from .dataset import Dataset
from .executable import Executable
from .externaljob import ExternalJob
from .instrument import Instrument
from .multiplexid import MultiplexId
from .multiplexkit import MultiplexKit
from .order import Order
from .parameter import Parameter
from .plate import Plate
from .project import Project
from .resource import Resource
from .run import Run
from .sample import Sample
from .storage import Storage
from .user import User
from .workflow import Workflow
from .workflowstep import WorkflowStep
from .workflowtemplate import WorkflowTemplate
from .workflowtemplatestep import WorkflowTemplateStep
from .workunit import Workunit

__all__ = [
    "Application",
    "Dataset",
    "Executable",
    "ExternalJob",
    "Instrument",
    "MultiplexId",
    "MultiplexKit",
    "Order",
    "Parameter",
    "Plate",
    "Project",
    "Resource",
    "Run",
    "Sample",
    "Storage",
    "User",
    "Workflow",
    "WorkflowStep",
    "WorkflowTemplate",
    "WorkflowTemplateStep",
    "Workunit",
]



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from bfabric.entities.core.entity import Entity
from bfabric.entities.core.has_many import HasMany
from bfabric.entities.core.user_created_mixin import UserCreatedMixin

if TYPE_CHECKING:
    from bfabric import Bfabric
    from bfabric.entities import WorkflowTemplateStep


class WorkflowTemplate(Entity, UserCreatedMixin):
    ENDPOINT = "workflowtemplate"

    def __init__(self, data_dict: dict[str, Any], client: Bfabric | None = None) -> None:
        super().__init__(data_dict=data_dict, client=client)

    workflow_template_steps: HasMany[WorkflowTemplateStep] = HasMany(
        entity="WorkflowTemplateStep", bfabric_field="workflowtemplatestep"
    )



from __future__ import annotations

from pathlib import Path
from typing import Any, TYPE_CHECKING

from bfabric.entities.core.entity import Entity
from bfabric.entities.core.has_one import HasOne

if TYPE_CHECKING:
    from bfabric import Bfabric
    from bfabric.entities.storage import Storage
    from bfabric.entities.workunit import Workunit
    from bfabric.entities import Sample


class Resource(Entity):
    ENDPOINT = "resource"

    def __init__(self, data_dict: dict[str, Any], client: Bfabric | None = None) -> None:
        super().__init__(data_dict=data_dict, client=client)

    storage: HasOne[Storage] = HasOne("Storage", bfabric_field="storage")
    workunit: HasOne[Workunit] = HasOne("Workunit", bfabric_field="workunit")
    sample: HasOne[Sample] = HasOne("Sample", bfabric_field="sample", optional=True)

    @property
    def storage_relative_path(self) -> Path:
        """Returns the relative path of the resource in the storage as a Path object."""
        return Path(self["relativepath"].lstrip("/"))

    @property
    def storage_absolute_path(self) -> Path:
        """Returns the absolute path of the resource in the storage as a Path object."""
        return Path(self.storage.base_path) / self.storage_relative_path

    @property
    def filename(self) -> str:
        """Returns the filename of the actual path, i.e. not necessarily the resource name but rather the name
        as the file is stored.
        """
        return self["relativepath"].split("/")[-1]



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from bfabric.entities.core.has_user import HasUser

if TYPE_CHECKING:
    from bfabric.entities.core.has_container_mixin import EntityProtocol


class UserCreatedMixin:
    @property
    def created_at(self: EntityProtocol) -> datetime.datetime:
        iso = self.data_dict["created"]
        return datetime.datetime.fromisoformat(iso.replace("Z", "+00:00"))

    @property
    def updated_at(self: EntityProtocol) -> datetime.datetime:
        iso = self.data_dict["updated"]
        return datetime.datetime.fromisoformat(iso.replace("Z", "+00:00"))

    created_by: HasUser = HasUser(bfabric_field="createdby")
    updated_by: HasUser = HasUser(bfabric_field="updatedby")
